use strict;
use warnings;

use FindBin '$Bin';
use lib "$Bin/lib";    
use Test::More;

use JSON;
use Slurp;
use Test::Rest;
use TestUtils;
use Data::Dumper;

BEGIN { use Catalyst::Test 'Bugzilla::API' }

my ($res, $req, $old, $new, $json);
my $t = Test::Rest->new('content_type' => 'application/json');
##############################################################################

my $bugid_normal   = "9";
my $bugid_defaults = "99"; # assignee and QA contact defaults
my $bugid_midair   = "102"; # mid-air collisions

my $oneway = slurp("$Bin/json/oneway.json");
my $otherway = slurp("$Bin/json/otherway.json");

$new = from_json($oneway, {utf8 => 1});

my @multis = (
  { 
    'blocks'     => $new->{'blocks'},
    'depends_on' => $new->{'depends_on'},
  },
  { 
    'status'     => $new->{'status'},
    'resolution' => $new->{'resolution'},
  },
  { 
    'product'          => $new->{'product'},
    'component'        => $new->{'component'},
    'version'          => $new->{'version'},
    'target_milestone' => $new->{'target_milestone'},
  },
);

my @exclude = ("id", "ref", "creation_time", "creator",
               "is_confirmed", "actual_time", "remaining_time",
               "blocks", "depends_on",
               "status", "resolution", 
               "version", "component", "product", "target_milestone");

# Work out how many tests there are...
my $put_and_check_bug_ct = $put_and_get_ct + 1;

my $whole_change_ct  = (2 * $put_and_check_bug_ct);
my $one_at_a_time_ct = ((keys(%$new) - @exclude) * $put_and_get_ct) +
                       (@multis * $put_and_get_ct) + 1 + $put_and_check_bug_ct;
my $set_default_ct   = (2 * $put_and_get_ct) + 2;
my $midair_ct        = 2;
my $idempotency_ct   = 1 + $put_and_get_ct + 1;

plan(tests => $whole_change_ct + $one_at_a_time_ct + $set_default_ct
              + $midair_ct + $idempotency_ct);

sub put_and_check_bug {
  my ($url, $json, $name) = @_;
  
  my $newjson = put_and_get($url, $json, $name);
  
  my $new = from_json($newjson, {utf8 => 1});
  my $old = from_json($json, {utf8 => 1});

  delete($old->{'last_change_time'});
  delete($new->{'last_change_time'});
  
  is_deeply(canonicalize($new), canonicalize($old), "Bug now set $name");
}

##############################################################################
# Change test - change twice to make sure there's a change
# Currently no changes to: attachments.
##############################################################################
my $url = "/bug/$bugid_normal?" . $LOGIN;

put_and_check_bug($url, $oneway, "one way");
put_and_check_bug($url, $otherway, "the other way");

##############################################################################
# Change test - change fields one at a time
##############################################################################
foreach my $field (keys %$new) {
  # Some fields can't be changed, or can't be changed singly
  next if grep(/^$field$/, @exclude);
  $json = to_json({ $field => $new->{$field} }, {utf8 => 1});
  put_and_get($url, $json, "setting single field: $field");
}

foreach my $multi (@multis) {
  $json = to_json($multi, {utf8 => 1});
  my @keys = keys(%$multi);
  put_and_get($url, $json, "setting " . $keys[0] . " etc.");
}

# Result should be the same as if we'd done them all at once
$res = request($url);
$new = from_json($res->decoded_content, {utf8 => 1});
$old = from_json($oneway, {utf8 => 1});

delete($old->{'last_change_time'});
delete($new->{'last_change_time'});

is_deeply(canonicalize($new), canonicalize($old), "Bug OK 1 at a time");

# Reset
put_and_check_bug($url, $otherway, "the other way from 1 at a time");

##############################################################################
# Set default assignee and QA contact (bug 99, as it's in a component which
# has them)
##############################################################################
my $set = slurp("$Bin/json/set-assignee-qa.json");
my $unset = slurp("$Bin/json/unset-assignee-qa.json");

$url = "/bug/$bugid_defaults?" . $LOGIN;

put_and_get($url, $set, "Setting Assigned To/QA Contact");
my $newjson = put_and_get($url, $unset, "Unsetting Assigned To/QA Contact");
$new = from_json($newjson, {utf8 => 1});

is($new->{'assigned_to'}->{'name'}, 'mrslate@gerv.net', "Assigned To correct");
is($new->{'qa_contact'}->{'name'}, 'mrslate@gerv.net', "QA Contact correct");

##############################################################################
# Create a mid-air
##############################################################################
$url = "/bug/$bugid_midair?" . $LOGIN;

my $token = get_token($url);
$new = from_json($otherway, {utf8 => 1});
$new->{'update_token'} = $token;
# Hack last_change_time to cause a mid-air
$new->{'last_change_time'} = "2009-11-17T00:00:00Z";
$json = to_json($new, {utf8 => 1});

$req = $t->put(url  => $url, 
               data => $json);

$res = request($req);
is($res->code, 400, "Error returned");
my $error = from_json($res->decoded_content, {utf8 => 1});
is($error->{'message'}, "Mid-air collision", "Message correct");

##############################################################################
# Idempotency test
##############################################################################
$url = "/bug/$bugid_normal?" . $LOGIN;

$res = request($url);
ok($res->is_success, "Idempotency GET should succeed");  
my $orig = from_json($res->decoded_content, {utf8 => 1});

$newjson = put_and_get($url, $res->decoded_content, "Idempotency");
$new = from_json($newjson, {utf8 => 1});

delete($orig->{'update_token'});
delete($new->{'update_token'});

is_deeply($new, $orig, "Idempotency test");

done_testing();
