use strict;
use warnings;

use FindBin '$Bin';
use lib "$Bin/lib";    
use Test::More;

use JSON;
use Slurp;
use Test::Rest;
use TestUtils;
use Data::Dumper;

BEGIN { use Catalyst::Test 'Bugzilla::API' }

my ($res, $req, $old, $new, $json);
my $t = Test::Rest->new('content_type' => 'application/json');
##############################################################################

my $bugid = 10;
my $attach4url = '/attachment/4?' . $LOGIN;

my $putandcheck_ct = 3;
plan(tests => 5 + (3 * $putandcheck_ct) + 2 + 2);

##############################################################################
# Idempotency test
##############################################################################
$res = request($attach4url);
ok($res->is_success, 'Request should succeed');
$json = $res->decoded_content;
my $orig = from_json($json, {utf8 => 1});

$req = $t->put(url  => $attach4url, 
               data => $json);

$res = request($req);
ok($res->is_success, 'Update should succeed');
my $okcheck = from_json($res->decoded_content, {utf8 => 1});
is($okcheck->{'ok'}, 1, "Result is OK");

$res = request($attach4url);
ok($res->is_success, 'Request should succeed');
$new = from_json($res->decoded_content, {utf8 => 1});

foreach my $field ('update_token', 'last_change_time') {
  delete($orig->{$field});
  delete($new->{$field});
}

is_deeply($new, $orig, "Idempotency test");

##############################################################################
# Change test for single fields - change twice to make sure there's a change
##############################################################################
sub putandcheck {
  my ($json, $name) = @_;

  my $token = get_token($attach4url);
  $json = add_field_to_json($json, 'update_token', $token);
  
  $req = $t->put(url  => $attach4url, 
                 data => $json);

  $res = request($req);
  ok($res->is_success, "$name (put) should succeed");
  
  $res = request($attach4url);
  ok($res->is_success, '$name (get 2) should succeed');
  $new = from_json($res->decoded_content, {utf8 => 1});
  $old = from_json($json, {utf8 => 1});

  foreach my $field ('update_token', 'last_change_time', 'comments') {
    delete($old->{$field});
    delete($new->{$field});
  }

  is_deeply(canonicalize($new), canonicalize($old), "Attachment ended OK");
}

my $oneway = slurp("$Bin/json/attachment-oneway.json");
my $otherway = slurp("$Bin/json/attachment-otherway.json");

putandcheck($oneway, "One way (single)");
putandcheck($otherway, "Other way (single)");

##############################################################################
# Add comment test
##############################################################################
my $random_number = rand();
my $comment = { "text" => "A comment with a random no.: $random_number" };
$otherway = add_field_to_json($otherway, 'comments', [$comment]);
putandcheck($otherway, "Other way (adding comment)");

$res = request("/bug/$bugid/comment?" . $LOGIN);

ok($res->is_success, 'Request should succeed');
$new = from_json($res->decoded_content, {utf8 => 1});
my $text = $new->{'comments'}->[-1]->{'text'};

like($text, qr/$random_number/, "Comment added OK with attachment update");

##############################################################################
# Create a mid-air
##############################################################################
my $token = get_token($attach4url);
$new = from_json($otherway, {utf8 => 1});
$new->{'update_token'} = $token;
# Hack last_change_time to cause a mid-air
$new->{'last_change_time'} = "2009-11-17T00:00:00Z";
$json = to_json($new, {utf8 => 1});

$req = $t->put(url  => $attach4url, 
               data => $json);

$res = request($req);
is($res->code, 400, "Error returned");
my $error = from_json($res->decoded_content, {utf8 => 1});
is($error->{'message'}, "Mid-air collision", "Message correct");

done_testing();
