use strict;
use warnings;

use FindBin '$Bin';
use lib "$Bin/lib";    
use Test::More;

use JSON;
use Slurp;
use Test::Rest;
use TestUtils;
use Data::Dumper;

BEGIN { use Catalyst::Test 'Bugzilla::API' }

my ($res, $req, $old, $new, $json);
my $t = Test::Rest->new('content_type' => 'application/json');
##############################################################################

# A list of URLs and the fields which they should remove
my %comments_urls = (
  "include_fields=_all" => [],
  "include_fields=_default" => [],
  "include_fields=is_private,text,creator,id" => [],
  "include_fields=text,creator" => ["is_private", "id"],
  "include_fields=is_private,id" => ["text", "creator"],
  "include_fields=id,creator" => ["text", "is_private"],
  "include_fields=text" => ["is_private", "creator", "id"],
  "exclude_fields=id" => ["id"],
  "exclude_fields=creator,text" => ["creator", "text"],
  "exclude_fields=creator, text, is_private" => ["creator", "text",
                                                                "is_private"],
  "include_fields=_default&exclude_fields=is_private" => ["is_private"],
  "include_fields=_all&exclude_fields=text" => ["text"],
  "include_fields=text,creator&exclude_fields=creator" => ["creator", "id",
                                                                "is_private"],
  "include_fields=text&exclude_fields=text" => ["text", "id", "creator",
                                                                "is_private"],
);

# Hash values are the fields you expect to see...
my %one_attachment_urls_positive = (
  "include_fields=description" => ["description"],
  "include_fields=description,size,file_name" => ["description", "size", 
                                                  "file_name"],    
  "include_fields=data" => ["data"],
  "include_fields=data,encoding" => ["data", "encoding"],
  "include_fields=data,encoding,description" => ["data", "encoding", 
                                                 "description"],
);

# Hash values are the fields you expect to have been removed.
my %one_attachment_urls_negative = (
  "include_fields=_all" => [],
  "include_fields=_default" => ["data", "encoding"],
  "include_fields=_default,data" => ["encoding"],
  "include_fields=_default,data,encoding" => [],
  "exclude_fields=description" => ["data", "encoding", "description"],
  "exclude_fields=description, size" => ["data", "encoding", 
                                         "description", "size"],
);

plan(tests => (keys(%comments_urls) * 2) +
              (keys(%one_attachment_urls_positive) * 2) +
              (keys(%one_attachment_urls_negative) * 2));

sub remove_fields {
  my ($arrayref, $fields) = @_;
  
  foreach my $obj (@$arrayref) {
    foreach my $field (@$fields) {
      delete $obj->{$field};
    }
  }
}

sub remove_other_fields {
  my ($arrayref, $fields) = @_;
  
  my %keepfields = map { $_ => 1 } @$fields;
  
  foreach my $obj (@$arrayref) {
    foreach my $field (keys %$obj) {
      if (!$keepfields{$field}) {
        delete $obj->{$field};
      }
    }
  }
}

##############################################################################
# Include/exclude on a fairly simple "default is all" call - using comments
##############################################################################
my $bugid = 1;

foreach my $query (keys %comments_urls) {
  $res = request("/bug/$bugid/comment?" . $query);

  ok($res->is_success, 'Request should succeed');
  $new = from_json($res->decoded_content, {utf8 => 1});

  $old = load("bug1comments.json");
  remove_fields($old->{'comments'}, $comments_urls{$query});

  # Not bothering with attachment stuff - number of fields gets too great
  # and fiddly to specify all the include/excludes. And solves the 3.6
  # compatibility issue too.
  remove_fields($old->{'comments'}, 
                        ["creation_time", "attachment_id", "attachment_ref"]);
  remove_fields($new->{'comments'},
                        ["creation_time", "attachment_id", "attachment_ref"]);
  
  is_deeply(canonicalize($new), canonicalize($old), 
                                         "Comments OK for incl/excl: $query");
}

##############################################################################
# Attachment list and single attachment just have two fields differing between
# _default and _all - data and encoding.
##############################################################################
my $aid = 2;

foreach my $query (keys %one_attachment_urls_positive) {
  $res = request("/attachment/$aid?" . $query);

  ok($res->is_success, 'Request should succeed');
  $new = from_json($res->decoded_content, {utf8 => 1});

  $old = load("attachment$aid-all.json");
  $old = canonicalize($old);

  remove_other_fields([$old], $one_attachment_urls_positive{$query});
  
  is_deeply(canonicalize($new), canonicalize($old),
                                        "Attachment OK for incl/excl: $query");
}

foreach my $query (keys %one_attachment_urls_negative) {
  $res = request("/attachment/$aid?" . $query);

  ok($res->is_success, 'Request should succeed');
  $new = from_json($res->decoded_content, {utf8 => 1});

  $old = load("attachment$aid-all.json");
  $old = canonicalize($old);
  remove_fields([$old], $one_attachment_urls_negative{$query});
  
  is_deeply(canonicalize($new), canonicalize($old),
                                        "Attachment OK for incl/excl: $query");
}

done_testing();
