# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Bugzilla HTTP REST API proxy.
#
# The Initial Developer of the Original Code is
# the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2009
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   The Original Software was written to the Glory of God by 
#   Gervase Markham <written.to.the.glory.of.God@gerv.net>.
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

package TestUtils;

use strict;
use warnings;
use base 'Exporter';

use Catalyst::Test 'Bugzilla::API';
use Test::More;

use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

@EXPORT = qw(
  $USERNAME
  $PASSWORD
  $LOGIN
  $STRIP_TIMES
  load
  add_field_to_json
  get_token
  $get_token_ct
  canonicalize
  put_and_get
  $put_and_get_ct
  put_and_check_flags
  $put_and_check_flags_ct
  get_bugzilla_version
);

$VERSION = '0.01';

use Data::Dumper;
use JSON;
use Slurp;
use FindBin '$Bin';

our $STRIP_TIMES = 0;

our $USERNAME = 'gerv@mozilla.org';
our $PASSWORD = 'gervgerv';

our $LOGIN = "username=$USERNAME&password=$PASSWORD";

sub load {
  my ($filename) = @_;
  my $data = from_json(slurp("$Bin/json/$filename"), {utf8 => 1});
  
  # Compatibility massaging
  my $version = get_bugzilla_version();
  if ($version < 4.0) {
    if ($filename =~ /bug1(-li)?\.json$/ && defined($data->{'bugs'})) {
      # In 4.0, Bugzilla started returning URL and multiply-valued custom 
      # fields on the CSV list interface.
      delete($data->{'bugs'}->[0]->{'url'});
      delete($data->{'bugs'}->[0]->{'cf_multiple'});
    }
  }
  
  return $data;
}

# Add an extra field on to the end of a JSON string
sub add_field_to_json {
  my ($json, $field, $value) = @_;
  $DB::single = 1;
  my $obj = from_json($json, {utf8 => 1});
  $obj->{$field} = $value;
  return to_json($obj,{utf8 => 1});
}

sub get_token {
  my ($url) = @_;
  
  my $res = request($url);
  my $obj = from_json($res->decoded_content, {utf8 => 1});
  return $obj->{'update_token'};
}

sub put_and_get {
  my ($url, $json, $name) = @_;
  if (ref($json)) {
    diag("Passed non-string to put_and_get");
    return undef;
  }
  
  my $token = get_token($url);
  $json = add_field_to_json($json, 'update_token', $token);
  
  my $t = Test::Rest->new('content_type' => 'application/json');
  
  my $req = $t->put(url  => $url, 
                    data => $json);

  my $res = request($req);
  ok($res->is_success, "Put OK for $name");
  my $okcheck = from_json($res->decoded_content, {utf8 => 1});
  is($okcheck->{'ok'}, 1, "PUT result OK for $name");
  
  $res = request($url);
  ok($res->is_success, "Get 2 OK for $name");

  return($res->decoded_content);
}

our $put_and_get_ct = 3;

sub put_and_check_flags {
  my ($url, $json, $name, $del_extra_fields) = @_;

  $del_extra_fields ||= 0;
  
  my $newjson = put_and_get($url, $json, $name);
  
  my $new = from_json($newjson, {utf8 => 1});
  my $old = from_json($json, {utf8 => 1});

  $old = { 'flags' => $old->{'flags'} };
  $new = { 'flags' => $new->{'flags'} };

  if ($del_extra_fields) {
    foreach my $obj ($old, $new) {
      foreach my $flag (@{$obj->{'flags'}}) {
        delete($flag->{'id'});
        delete($flag->{'setter'});
      }
    }
  }
  
  is_deeply(canonicalize($new), canonicalize($old), "Flags correct for $name");
}

our $put_and_check_flags_ct = $put_and_get_ct + 1;

# Make a bug or other object suitable for an is_deeply comparison by 
# removing all the bits which vary from run to run.
# Safe for: bugs, attachments
sub canonicalize {
  my ($obj) = @_;

  delete($obj->{'update_token'});

  maybe_strip_time($obj, 'creation_time');
  maybe_strip_time($obj, 'last_change_time');
  
  foreach my $ref ('ref', 'bug_ref') {
    strip_port($obj, $ref);
  }
  
  foreach my $person ('creator', 'assigned_to', 'qa_contact', 'attacher') {
    strip_port($obj->{$person}, 'ref');
  }
  
  if ($obj->{'attachments'}) {
    foreach my $obj (@{$obj->{'attachments'}}) {
      delete($obj->{'update_token'});
      maybe_strip_time($obj, 'creation_time');
      maybe_strip_time($obj, 'last_change_time');
      strip_port($obj->{'attacher'}, 'ref');
      strip_port($obj, 'bug_ref');
      strip_port($obj, 'ref');
      
      canonicalize_flags($obj);
    }
  }

  canonicalize_flags($obj);

  if ($obj->{'cc'}) {
    foreach my $obj (@{$obj->{'cc'}}) {
      strip_port($obj, 'ref');
    }
  }
  
  if ($obj->{'comments'}) {
    my $version = get_bugzilla_version();
    
    foreach my $obj2 (@{$obj->{'comments'}}) {
      strip_port($obj2->{'creator'}, 'ref');
      strip_port($obj2, 'attachment_ref');
      maybe_strip_time($obj2, 'creation_time');
      
      if ($version <= 3.4) {
        delete($obj2->{'attachment_id'});
        delete($obj2->{'attachment_ref'});
        # Auto-generated text for attachment creation changed
        if ($obj2->{'text'}) {
          $obj2->{'text'} =~ 
                   s/Created an attachment \(id=(\d+)\)/Created attachment $1/;
        }
      }
    }
  }

  if ($obj->{'history'}) {
    foreach my $obj (@{$obj->{'history'}}) {
      strip_port($obj->{'changer'}, 'ref');
      maybe_strip_time($obj, 'change_time');
    }
  }
  
  return($obj);
}

sub canonicalize_flags {
  my ($obj) = @_;
  
  if ($obj->{'flags'}) {
    foreach my $flag (@{$obj->{'flags'}}) {
      strip_port($flag->{'setter'}, 'ref');
      strip_port($flag->{'requestee'}, 'ref') if $flag->{'requestee'};
    }
  }
}

# Using the API via the net puts port numbers on the ref URLs; using it via
# the tests does not. This sub means we can just paste from the net version
# into reference files and the comparisons will pass.
sub strip_port {
  my ($obj, $field) = @_;
  $obj->{$field} =~ s/:3000// if $obj->{$field};
}

sub maybe_strip_time {
  my ($obj, $field) = @_;
  delete($obj->{$field}) if $STRIP_TIMES;
}

sub get_bugzilla_version {
  my $config = slurp("$Bin/../bugzilla_api.conf");
  $config =~ /\n\s*bugzilla_version\s+(\d\.\d)/s;
  my $version = $1;
  return $1;
}

1;
